#!/usr/bin/perl


# Contemplate Assembler (assembler.cgi) version 1.2.4
# Copyright 1999-2008 Type A Learning Agency (http://www.typea.net)
# This file may be used, modified, or redistributed under the terms
# of the open-source agreement contained in docs/license.txt


# use the CGI.pm module to access CGI functionality

	use CGI;
	$cgi = new CGI;


# if the script is being used by the flattener or formulator, we'll give a different kind of output

	$skip_rendering = $cgi->param('skip_rendering');
	$formulator_fill = $cgi->param('formulator_fill');
	$settings_used = "skip_rendering,formulator_fill,"; # this prevents these values from being changed into arrays like other arguments
	

# set visible text here for easier localization

	$search_label_success = "Search results for";
	$search_label_failure = "Can't find";
	$search_label_none = "Please enter a search string";
	
	
# initialize variables that we might use before setting

	$error_log = "";
	$show_errors = "";
	$output = "";
	$output_two = "";
	$pre_render_files = 0;
	$read_page_definitions_from = "";
	$page_definitions = "";
	$offset = 0;
	@embed_variable_keys = ();
	@embed_variable_values = ();
	@embeds = ();
	@tag_names = ();
	$tag_names_flattened = "";
	$file_fields = "";
	$stored_selections = "";
	$latest_mod_date = 0;	
	$latest_mod_date_formatted = "";	
	$visitor_ID = "";
	$template = "";
	$page_definition = "";
	$header = "";
	

# define some useful subroutines

	sub get_file {
		my $file = $_[0]; my $method = $_[1]; my $check_last_modified = $_[2]; my $text = ""; # method can be "http", "filesystem", or "auto" (auto looks at $pre_render_files list)
		if (($file =~ /\/(\.\w.*)/)) { # don't read files whose names start with "."
			$text = show_error("Not authorized to read the hidden file named \"$1\"");
			$file_missing = 1;
		} elsif (($file =~ /[^\w-\.\/\?\&\=\%]/)) { # don't read files whose names contain characters that could be used to read remote files, or to hack into the open command (Perl only)
			$text = show_error("File names can't contain the character $&");
			$file_missing = 1;
		} else {
			
			# first, figure out if we're going to read over HTTP or through the filesystem...
			if (($method eq "")||($method eq "auto")) {
				if (@pre_render_files) {
					for ($n=0; $n < @pre_render_files; $n++) {
						if ($file =~ /@pre_render_files[$n]/) { $method = "http"; }
					}
				}
			}
			
			if ($method eq "http") {
				eval 'use LWP::Simple'; 
				if ($@) {
					$output .= show_error("Can\'t access Perl module LWP::Simple for pre-rendering files<br>Reading files directly instead<br>Hint: Set pre-render files to 0 in preferences.ini to skip this routine");
					$method = "filesystem";
				}
			}
			
			# ...then get the file
			if ($method eq "http") {
			
				# 1) prepare the URL for this type of connection
				$file_URL = absolute_path($file, "http");

				# 2) read the file
				$myfile = get($file_URL);
				$file_exists = 1; # to do: how can we test for this?
				if (! $file_exists) {
					$text = show_error("Can't open the file named \"" + $file + "\"" + $check_URL);
					$file_missing = 1;
				} else {
					$text = $myfile;
				}
				
				#$output .= "<p>reading $file over http<p>";  # debug
				
			} else {
			
				# 1) prepare the URL for this type of connection
				$file_URL = absolute_path($file, "filesystem");
				
				# 2) open a connection to the file
				open (myfile, $file_URL) or
				$text = show_error("Can't open the file named \"$file\"<br>$! $check_URL");
				if ($text =~ /Assembler error/) {
					$file_missing = 1;
				} else {
				
					# 3) read the file
					while (<myfile>) { $text .= "$_"; }
					$file_missing = 0;
					
					# 4) for local files, catch the file's modification date (for $read_page_definitions_from and $template, plus content files when the template contains a env_last_modified tag)
					if ($check_last_modified) {
						($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size, $atime, $mtime, $ctime, $blksize, $blocks) = stat($file_URL);
						if ($mtime > $latest_mod_date) {
							$latest_mod_date = $mtime;
							($second, $minute, $hour, $day, $month, $year, $weekday, $yearday, $dst_flag) = localtime($latest_mod_date);
							$year += 1900;
							$month = sprintf("%02d",$month+1);
							$day = sprintf("%02d",$day);
							if ($hour < 12) {
								$ampm = "am";
							} else {
								$ampm = "pm";
								if ($hour > 12) {
									$hour = $hour - 12;
								}
							}
							#$hour = sprintf("%02d",$hour);
							$minute = sprintf("%02d",$minute);
							$latest_mod_date_formatted = "$year-$month-$day - $hour:$minute $ampm";
						}
					}

					# 5) close the file connection
					close (myfile);
					
				}
				
				#$output .= "<p>reading $script_location_copy + $file_copy through filesystem<p>";

			}
			
			# finally, conform all line breaks to avoid cross-platform issues
			# we'd rather conform to Unix line breaks (\n), but that doesn't work in ASP, which doesn't support the s flag in regular expressions
			$text =~ s/\r\n/\r/g; # change DOS to Mac
			$text =~ s/\n/\r/g; # change Unix to Mac
	
		}
		return $text;
	}
	
	sub get_field {
		my $source_text = $_[0]; my $field_name = $_[1]; my $file_type = $_[2];
		if ($file_type =~ /xml/i) {
			
			$source_text = make_XML_paths($source_text);
			$tag_type = "content";
			if ($source_text =~ /<$tag_type[^>]+name=["']?$field_name\b[^>]*>(.*?)<\/$tag_type>/is) {
				$field_text = $1;
			} else {
				$field_text = show_error("Can't find the $tag_type tag named \"$field_name\" in $content_file $check_URL");
			}

		} else {		

			unless ($field_name =~ /$field_marker/) {  # but what if formatting gets between the marker and the label?
				$field_name = $field_marker.$field_name;  # in that case, remove these lines and change $content_field to $field_marker.*$content_field below
			}
			if ($source_text =~ /$field_name(\W.*\/td>.*<td[^>]*>.*<\/td>)/is) {
				$field_text = $1;  # do a rough trim
				$field_text =~ s/<\/td>//i;  # remove up to the previous </td>
				$field_text =~ s/<td[^>]*>/split-here-hey/is;  # mark the opening <td>
				$field_text =~ s/$field_marker/split-here-hey/i;  # mark the next field marker
				@temp = split("split-here-hey", $field_text);  # trim at the markers
				$field_text = @temp[1];
				if ($field_text =~ /(.*)<\/td>/is) {  # remove from the trailing </td>
					$field_text = $1
				}
			} else {
				unless ($file_missing) {
					$field_text = show_error("Can't find the field named \"$field_name\" in $content_file $check_URL $check_field_label");
				}
			}

		}
		return $field_text;
	}
	
	sub get_form {
		my $source_text = $_[0]; my $form_element = $_[1];
		$value = "";
		
		# this function doesn't recognize values like this: <input name = foo> we could make it do so by adding /s* around all the equals signs, but it doesn't now
		
		# before looking for the value, let's find out if it has single or double quotes around it or not
		if ($source_text =~ /name=[\"\']?$form_element\W[^>]*value=[^\"\']/is) { $value_match = "([^ >]*)"; }
		elsif ($source_text =~ /name=[\"\']?$form_element\W[^>]*>.*?<option[^>]+value=[^\"\']/is) { $value_match = "([^ >]*)"; } # the previous line covers all other input types, but the select menu is different because the type and value are in different tags (we wouldn't need this if we could write a regex that spanned the <select> and <option> tags, but that wasn't too greedy with other input types)
		elsif ($source_text =~ /name=[\"\']?$form_element\W[^>]*value=\'/is) { $value_match = "'([^\']*?)'"; }
		elsif ($source_text =~ /name=[\"\']?$form_element\W[^>]*>.*?<option[^>]+value=\'/is) { $value_match = "'([^\']*?)'"; } # the previous line covers all other input types, but the select menu is different because the type and value are in different tags (we wouldn't need this if we could write a regex that spanned the <select> and <option> tags, but that wasn't too greedy with other input types)
		else { $value_match = "\"([^\"]*?)\""; } # we'll let the double quote be the default, so if something goes wrong we still might get the desired result
 
		# okay, now let's look for it
		if ($source_text =~ /<input[^>]+name=["']?$form_element\W[^>]*type=["']?text["']?[^>]+value=$value_match[^>]*>/is) { $value = $1; }
		elsif ($source_text =~ /<input[^>]+name=["']?$form_element\W[^>]*type=["']?radio["']?[^>]+value=$value_match[^>]+checked[^>]*>/is) { $value = $1; }
		elsif ($source_text =~ /<input[^>]+name=["']?$form_element\W[^>]*type=["']?checkbox["']?[^>]+value=$value_match[^>]+checked[^>]*>/is) {
			# return a list (not array) of all the checked boxes
			while ($source_text =~ /<input[^>]+name=["']?$form_element\W[^>]*type=["']?checkbox["']?[^>]+value=$value_match[^>]+checked[^>]*>/is) {
				$value .= "$1, ";
				$source_text =~ s/<input[^>]+name=["']?$form_element\W[^>]*type=["']?checkbox["']?[^>]+value=$value_match[^>]+checked[^>]*>//is;
			}
			$value = substr($value, 0, -2);
		}
		elsif ($source_text =~ /<select[^>]+name=["']?$form_element\W[^>]*>.*?<option[^>]+value=$value_match[^>]+selected[^>]*>.*?<\/select>/is) { $value = $1; }
		elsif ($source_text =~ /<textarea[^>]+name=["']?$form_element\W[^>]*>(.*?)<\/textarea>/is) { $value = $1; }
		else {
			if ($source_text =~ /name=[\"\']?$form_element\W/is) {
				# some HTML editors might remove the value attribute if it's empty, so maybe that's why we can't find the form element
				$value = "";
			} else {
				# otherwise, we give up
				$value = show_error("Can't find the form element named \"$form_element\" in $content_file $check_URL");
			}
		}
		return $value;
	}
	
	sub get_setting {
		my $source_text = $_[0]; my $setting = $_[1]; my $allow_override = $_[2]; my $allow_partial = $_[3]; my $value = "";
		
		# allow most preferences to be overridden through the argument string
		$setting_no_spaces = $setting;
		$setting_no_spaces =~ s/[ -]/_/g;
		$settings_used .= "$setting_no_spaces,";

		$override = $cgi->param($setting_no_spaces);
		if (($allow_override)&&($override ne "")) {
			$value = $override;
			# don't allow users to pass code through the URL arguments
			$value =~ s/</&lt;/i;
			$value =~ s/>/&gt;/i;
		} else {
			$source_text = "\r".$source_text;  # this makes it easy to make sure settings start on a new line
			if ($allow_partial) {  # this lets us use more verbose labels in the ini file
				if ($source_text =~ /[^\w]$setting[^:\r]*: ([^\r]*)\r*/i) {
					$value = $1;
				}
			} else {  # this ensures an exact match in the pages file
				if ($source_text =~ /[^\w]$setting: ([^\r]*)\r*/i) {
					$value = $1;
				}
			}
		}
		if ($value eq "0") { $value = 0; } # this ensures that a 0 value will evaluate false in every language
		return $value;
	}
	
	sub get_page_name { # look for a page definition that exactly matches a given dynamic URL
		my $page_definition = $_[0]; my $source = $_[1];
        $source = "\r".$source; # this makes it easier to match a page definition on the first line of the file
		$page_name = ($source =~ /[^\w]([-\w]+)\:\s+\Q$page_definition\E[\&\s\r]/) ? $1 : "" ;
		return $page_name;
	}
	
	sub get_page_names_containing { # look for page definitions that contain a given set of arguments
		my $args = $_[0]; my $source = $_[1];
		@pages = ();
		$source = "\r".$source; # this makes it easier to match a page definition on the first line of the file
		while ($source =~ /[^\w]([-\w]+)\:\s+[^\r]+=\Q$args\E[^\r]*[\&\s\r]/) {
			push(@pages, $1);
			$source =~ s/([^\w])[-\w]+\:\s+[^\r]+=\Q$args\E[^\r]*[\&\s\r]/$1/;
		}
		return @pages;
	}
	
	sub get_page_names { # return a list of all the page names in a page definitions file
		my $source = $_[0];
		@page_names = ();
		while ($source =~ /([-\w]+)\:\s+[\w]+/) {
			$this_page_name = $1;
			push(@page_names, $this_page_name);
			$source =~ s/$this_page_name\://;
		}
		return @page_names;
	}
	
	sub get_field_names { # return a list of all the content fields in a content file -- either HTML or XML depending on the second argument
		my $source = $_[0]; my $file_type = $_[1];
		@field_names = ();
		if ($file_type =~ /xml/i) {

			$source = make_XML_paths($source);
			$tag_type = "content";
			while ($source =~ /<$tag_type[^>]+name=["']?([-\w\/]+)["']?[^>]*>/is) {
				$this_field_name = $1;
				push(@field_names, $this_field_name);
				$source =~ s/<$tag_type[^>]+name=["']?[-\w\/]+["']?[^>]*>//is;
			}

		} else {

			while ($source =~ /$field_marker([-\w]+)/) {
				$this_field_name = $1;
				push(@field_names, $this_field_name);
				$source =~ s/$field_marker//;
			}

		}
		return @field_names;
	}
	
	sub get_form_names { # return a list of all the form elements in a content field
		my $source = $_[0];
		@form_names = ();
		while ($source =~ /<input[^>]+name=[\"\']?([-\w]+)\W[^>]*>/is) {
			$this_form_name = $1;
			push(@form_names, $this_form_name);
			$source =~ s/\Q$&\E//;
		}
		while ($source =~ /<select[^>]+name=[\"\']?([-\w]+)\W[^>]*>/is) {
			$this_form_name = $1;
			push(@form_names, $this_form_name);
			$source =~ s/\Q$&\E//;
		}
		while ($source =~ /<textarea[^>]+name=[\"\']?([-\w]+)\W[^>]*>/is) {
			$this_form_name = $1;
			push(@form_names, $this_form_name);
			$source =~ s/\Q$&\E//;
		}
		return @form_names;
	}
	
	sub make_XML_paths { # rewrite nested XML content as flat content with paths for tag names; assumes no more than one tag per line
		my $source = $_[0];
		@source = split("\r", $source);
		@path = ();
		for ($n=0; $n < @source; $n++) {
			$this_line = @source[$n];
			if ($this_line =~ /<group[^>]+name=["']?([-\w]+)["']?[^>]*>/is) {
				$group_name = $1;
				push(@path, $group_name); # add the new group name to the path array
			} elsif ($this_line =~ /<\/group>/i) {
				$group_name = pop @path; # remove the latest group name from the path array
			} elsif ($this_line =~ /<content[^>]+name=["']?([-\w]+)["']?[^>]*>/is) {
				$field_name = $1;
				if (@path > 0) {
					$current_path = join ("/", @path);
					$field_name = $current_path."/".$field_name;
					$this_line =~ s/(<content[^>]+name=["']?)[-\w]+(["']?[^>]*>)/$1$field_name$2/is;
					@source[$n] = $this_line;
				}
			}
		}
		$source = join("\r", @source);
		return $source;
	}
	
	sub get_directory_listing { # return a list of all the items in the specified directory
		my $directory = $_[0];
		$success = 1;
		opendir(dir, absolute_path($directory, "filesystem")) or $success = 0;
		if ($success) {
			#@files = grep readdir(dir); # that gets all files
			#@files = grep {/\w+\..\w+/} readdir(dir); # that gets files only
			@files = grep {/^[^\.]/} readdir(dir); # that gets files that don't start with a dot
			closedir dir;
		}
		return @files;
	}
	
	sub get_recursive_listing { # return a list of all the files in a given folder, including subfolders
		my $start_folder = $_[0]; my $recursive = $_[1]; my $filter = $_[2];
		$start_folder =~ s/[\/\\]$//; # remove a trailing slash
		@file_list = ();
		@folder_list = $start_folder;
		while (@folder_list) {
			$this_folder_path = @folder_list[0];
			@this_folder_items = get_directory_listing($this_folder_path);
			for ($n=0; $n<@this_folder_items; $n++) {
				$this_item_name = @this_folder_items[$n];
				$this_item_path = $this_folder_path."/".$this_item_name;
 				# if the item name contains a dot, we assume it's a file, if not, we assume it's a directory
 				if ($this_item_name !~ /\./) {
 					# we've found a folder; add it to the list of folders to scan
 					if ($recursive) {
						push(@folder_list, $this_item_path);
					}
				} elsif ((!$filter)||($this_item_name =~ /$filter/)) {
					# we've found a file; add it to the list of files to return
					$this_item_path =~ s/$start_folder\///; # make results relative to the start folder, not to the site
					push(@file_list, $this_item_path);
				}
			}
			shift @folder_list; # remove the folder we just scanned from the list
		}
		return @file_list;
	}
	
	sub process_tag_nav {
		my $this_type = $_[0];
					
		# if we haven't already loaded the page definitions file, do that now
		if (! $page_definitions) { $page_definitions = get_file(@read_page_definitions_from[0], "filesystem", 1); }
		
		# if we came here from the expanded URL and not the page shortcut, get the page name now
		if (! $page) { $page = get_page_name($page_definition, $page_definitions); }

		$nav_page = $page;
		
		@nav_pages = ();
		$nav_page_position = 0;
		$count_pages = 0;
		
		# find the block of pages (separated by double line breaks) that this page belongs to
		# first, split the page definitions file up into blocks
		$pages_blocks = "\r\r$page_definitions\r\r";
		$pages_blocks =~ s/\r\r/\r\r\r/g;
		if ($pages_blocks =~ /.*\r\r(.*?\W$nav_page: .*?)\r\r.*/) {
			# then find the block that contains this page
			$nav_block = $1;
			$nav_block =~ s/\:[ ]*[^\r]*//g;
			$nav_block =~ s/^\r*(.*)\r*$/$1/g; # no trim function?
			@nav_pages = split ("\r", $nav_block);
			$count_pages = @nav_pages;
		}
		for ($k=0; $k<$count_pages; $k++) {
			# finally, parse through it to find this page's position in the block
			if (lc(@nav_pages[$k]) eq lc($nav_page)) {
				$nav_page_position = $k+1;
				last; # stop the loop
			}
		}
					
		if ($this_type =~ /nav_current/) {
			$content = int($nav_page_position); # return the position of the current page
		} elsif ($this_type =~ /nav_total/) {
			$content = $count_pages; # return the number of pages in this block
		} elsif ($this_type =~ /nav_first/) {
			$content = @nav_pages[0]; # return the name of the first page
		} elsif ($this_type =~ /nav_last/) {
			$content = @nav_pages[$count_pages - 1]; # return the name of the last page
		
		} else {

			# don't bother with the next part, if the current page doesn't exist
			if ($nav_page_position > 0) {
		
				if ($this_type =~ /nav_next/) {
					$result_page_position = $nav_page_position + 1;
					if ($result_page_position > $count_pages) { $result_page_position = 1; }
					$content = @nav_pages[$result_page_position - 1]; # return the name of the requested page
				} elsif ($this_type =~ /nav_previous/) {
					$result_page_position = $nav_page_position - 1;
					if ($result_page_position < 1) { $result_page_position = $count_pages; }
					$content = @nav_pages[$result_page_position - 1]; # return the name of the requested page
				}
				
			} else {
				$content = show_error("Can't find the page named \"$nav_page\" in @read_page_definitions_from[0]");
			}
			
		}
		
		return $content;					
	}

	sub process_tag_env {
		my $this_type = $_[0];
		$this_type =~ /env_(.+)/;
		$variable = $1;
		if ($variable =~ "last_mod") {
			$content = $latest_mod_date_formatted;
		} else {
			$content = $ENV{uc($1)};
		}
		return $content;
	}

	sub show_error {
		my $message = $_[0];
		
		# prepare the shown message by removing quote marks
		$message_show = $message;
		$message_show =~ s/\"/&quot;/g;
		$message_show =~ s/'/&#039;/g;
		$message_show = "<table cellpadding=\"4\" border=\"1\"><tr><td class=\"error-text\"><b>Assembler error</b><br>$message_show</td></tr></table>";
		
		# prepare the hidden message by removing <br> tags
		$message_hide = $message_show;
		$message_hide =~ s/<br>/;/g;
		$message_hide = "<!-- Assembler error: $message_hide -->";
		
		# prepare the logged message by removing all but the first line
		$message_log = $message;
		if ($message_log =~ /<br>/) {
			$message_log = substr($message_log, 0, index($message_log, "<br>"));
		}
		$message_log = strip_tags($message_log); # also remove any HTML formatting from it
		
		# add the log message to the ongoing error log
		if ($error_log) { $error_log .= ","; } # and comma-delimit multiple errors on a page
		$error_log .= $message_log;
		
		# add the shown and hidden messages to the output
		if (($show_errors)&&($formulator_fill ne "-1")) {
			# only display one error at a time, since subsequent errors are usually a result of the first one
			if ($output !~ /Assembler error/) {
				return $message_show;
			} else {
				return "";
			}
		} else {
			return $message_hide;
		}
		
	}
	
	sub relative_path { # takes a reference to a file within our site and changes the initial delimiters if necessary
		my $path = $_[0]; my $subfolder = $_[1]; $this_read_file = 0; $authenticated = 1; $error = ""; $validate_path = 0;
		$path =~ s/\:\/\///g;  # don't allow files from other servers
		if ($path =~ /\//) { $validate_path = 1; } # we can skip the validation if we just have a file directly the content folder
		if ((substr($path,0,3) ne "../")&&(substr($path,0,2) ne "./")) {
			# if the path doesn't include a site subfolder, add the subfolder we passed in
			if ($subfolder) { $subfolder .= "/"; }
			$path = "../".$subfolder.$path;
		}
		if ($validate_path) {
			# allow for nonstandard paths, if they're listed in the ini file
			if (@read_files_from) {
				for ($n=0; $n < $read_files_from_count; $n++) {
					if ($read_files_recursively) {
						# allow matches to subfolders
						if ($path =~ /^@read_files_from[$n].*/) { $this_read_file = 1; }
					} else {
						# require an exact match to this folder
						$path_no_filename = $path;
						$path_no_filename =~ s/(.*\/)[^\/]*/$1/;
						if ($path_no_filename eq @read_files_from[$n]) { $this_read_file = 1; }
					}
				}
			}
			#$output .= "<p>this_read_file: $this_read_file<p>";  # debug
			if (! $this_read_file) {
				if ($path =~ /(.*\/)[^\/]*/i) { $path_trimmed = $1; }
				$error = show_error("Not authorized to read from the path \"$path_trimmed\"<br>Hint: Add nonstandard paths to \"read files from\" list in preferences.ini");
			}
		}
		if ((! $error)&&($require_authentication)) {
			# see if we require authentication to view this page and show an error if necessary
			$authenticated = 1;
			for ($n=0; $n<$require_authentication_count; $n++) {
				if ($path =~ /^@require_authentication[$n].*/) {
					$authenticated = ($login) ? 1 : 0 ;
					break;
				}
			}
			if (! $authenticated) {
				if ($path =~ /(.*\/)[^\/]*/i) { $path_trimmed = $1; }
				$error = show_error("Not authorized to read from the path \"$path_trimmed\"<br>Hint: Log in to \"require authorization to read files from\" list in preferences.ini");
				#$header = "Status: 403 Not Authorized\r"; # how can we make this work, along with the 404 header above?
			}
		}
		if ($error) {
			return $error;
		} else {
			return $path;
		}
	}
	
	sub absolute_path { # convert a reference to a file within our site to a complete filesystem path or absolute URL
		my $path = $_[0]; my $method = $_[1];
		
		if ($method eq "filesystem") {
		
			if ($script_location eq "NA") {
				$output .= show_error("Couldn't determine the filesystem path to this script. Please ensure that either the SCRIPT_FILENAME or PATH_TRANSLATED variable is available on your server.");
			} else {
				$script_location_copy = $script_location;
				$path_copy = $path;
				while ($path_copy =~ /\.\.\//) {
					# this avoids results like contemplate/../templates/default.html
					$script_location_copy =~ s/[^\/]+\/$//; # remove the last directory name from the script location
					$path_copy =~ s/\.\.\///; # remove the first updelim from the filename
				}
				$path = $script_location_copy.$path_copy;
			}

		} else {
		
			if ($request eq "NA") {
				$output .= show_error("Couldn't determine the web path to this script. Please ensure that either the SCRIPT_NAME or REQUEST_URI variable is available on your server.");
			} else {
				$thisserver = $ENV{"HTTP_HOST"};
				$thispath = "";
				if ($URL_rewriting_flag) {
					if ($request =~ /\/([^=]*\/)$URL_rewriting_flag\//i) { $thispath = $1; }
				} else {
					if ($request =~ /\/([^=]*\/)contemplate\//i) { $thispath = $1; }
				}
				
				# to do: how can we read the authentication info from this request so we can pass it on?
				$login = ""; 
				
				$path =~ s/\.\.\//http:\/\/$login$thisserver\/$thispath/i; # replace the first updelim with the scheme and path
				$path =~ s/\.\.\///g; # remove additional updelims
			}
		
		}
		
		return $path;
	}

	sub parse_int { # remove all but numeric characters from a string and convert to an integer
		my $str = $_[0];
		$str =~ s/\D//g;
		$int = int($str);
		return $int;
	}

	sub make_preview { # return a little bit before and a little bit after the selected text (used for search results)
		my $before_str = $_[0]; my $str = $_[1]; my $after_str = $_[2];
		$ellipses_before = (length($before_str) > $search_results_preview_length) ? "..." : "" ;
		$ellipses_after = (length($after_str) > $search_results_preview_length) ? "..." : "" ;
		$before_str = substr($before_str, -(int($search_results_preview_length / 2)));
		$after_str = substr($after_str, 0, int($search_results_preview_length / 2));
		$before_str =~ s/^\s+//;
		$after_str =~ s/\s+$//;
		$preview = $ellipses_before.$before_str."<b>$str</b>".$after_str.$ellipses_after;
		return $preview;
	}

	sub strip_tags { # remove all HTML or other markup from text; this is a substitute for PHP's strip_tags() built-in function
		my $str = $_[0];
		$str =~ s/>[\r\s]</></g;
		$str =~ s/<[^>]*>//g;
		$str =~ s/^\s*//;
		$str =~ s/\s*$//;
		return $str;
	}

	sub join_recursive { # if $this_element is an array, it comes through like ARRAY(0x39090239) and we can't do anything with it, so this function is useless
		my $delimiter = $_[0];
		$result = "";
		for ($n=1; $n<@_; $n++) {
			$this_element = $_[$n];
			#$output .= "join_recursive is checking $this_element<br>"; # debug
			if (is_array($this_element)) {
				#$output .= "$this_element is an array<br>"; # debug
				$result .= join_recursive($delimiter, $this_element);
			} else {
				$result .= $delimiter.$this_element;
			}
		}
		return $result;
	}
	
	sub is_array {
		my @array = $_[0];
		return (@array eq join("", @array)) ? 1 : 0 ;
		#return (ref(@array) eq "ARRAY") ? 1 : 0 ;
	}
	
	sub urlencode {
		my $string = $_[0];
		$string =~ s/([^A-Za-z0-9])/sprintf("%%%02X", ord($1))/seg;
		return $string;
	}

	sub urldecode {
		my $string = $_[0];
		$string =~ s/\%([A-Fa-f0-9]{2})/pack('C', hex($1))/seg;
		return $string;
	}

	sub random_string {
		my $strlen = $_[0];
		$chars = "0123456789bcdfghjklmnpqrstvwxyzBCDFGHJKLMNPQRSTVWXYZ";
		$string = "";
		for ($n=0; $n<$strlen; $n++) {
			$string .= substr($chars, int(rand(length($chars))), 1);
		}
		return $string;
	}


# set up some error messages that we can use later

	$check_URL = "<br>Hint: Check your page definition";
	$check_embed = "<br>Hint: Check your embed tag";
	$check_field_label = "<br>Hint: Check your content file and remove any formatting tags between the field marker and the field label";
	$change_tag_name = "<br>Hint: Choose a different name for this embed tag";
	

# set up some primary variables

	# try to figure out the path to this script as it appears in the browser, before any URL rewriting
	# in IIS, this probably won't work, because REQUEST_URI will either not exist, or will show the path after rewriting, since rewriting is handled externally from IIS
	$request = "NA";
	if ($ENV{'REQUEST_URI'}) {
		$request = $ENV{'REQUEST_URI'};
		if (index($request, "?") != -1) {
			$request = substr($request, 0, index($request, "?"));
		}
	} else {
		# IIS doesn't always provide the REQUEST_URI variable
		if ($ENV{'SCRIPT_NAME'}) {
			$request = $ENV{'SCRIPT_NAME'};
		}
	}

	# if we know the browser path, figure out the part that triggers the rewrite; assume this is just one virtual directory name
	$URL_rewriting_flag = "";
	if ($request !~ /contemplate\/assembler/) {
		if ($request =~ /([^\/]*)\/[^\/]+$/i) {
			$URL_rewriting_flag = $1;
		}
	}

	# figure out the full local path of the contemplate directory so we can build an absolute path that works on Windows computers
	# on Unix, this will include any symlinks used to construct the URL, so we can run the script from a symlink to the real contemplate folder, 
	# and paths will be relative to the apparent, rather than the actual, location of assembler.asp
	# on Windows, this will not include virtual directories, but constructing a path from a virtual directory wouldn't work anyway
	# because those aren't available at the file system level
	$script_location = "NA";
	if ($ENV{'SCRIPT_FILENAME'}) {
		$script_location = $ENV{'SCRIPT_FILENAME'};
	} else {
		# Windows doesn't provide the SCRIPT_FILENAME variable
		if ($ENV{'PATH_TRANSLATED'}) {
			$script_location = $ENV{'PATH_TRANSLATED'};
		}
	}
	$script_location =~ s/\\/\//g;
	if (($URL_rewriting_flag)&&($script_location =~ /(.*\/)$URL_rewriting_flag\//i)) {
		$script_location = $1."contemplate/";
	} else {
		$script_location =~ s/(.*\/)[^\/]+/$1/;
	}


# read in a preferences (.ini) file if one exists, and show which preferences may be overridden from the URL string

	$preferences = get_file("data/preferences.ini", "filesystem", 0);
	if ($preferences) {
		$field_marker = get_setting($preferences, "field marker", 1, 1);
		$log_traffic = get_setting($preferences, "log site traffic", 1, 1);
		$show_errors = get_setting($preferences, "show errors", 1, 1);
		$pre_render_files = get_setting($preferences, "pre-render files", 1, 1);
		$read_page_definitions_from = get_setting($preferences, "read page definitions from", 1, 1);
		$read_files_from = get_setting($preferences, "read files from", 0, 1);
		$read_files_recursively = get_setting($preferences, "read files recursively", 0, 1);
		$require_authentication = get_setting($preferences, "require authentication", 0, 1);
		$prevent_page_caching = get_setting($preferences, "prevent page caching", 1, 1);
		$show_expanded_URLs = get_setting($preferences, "show expanded URLs", 1, 1);
		$split_page_names_on = get_setting($preferences, "split page names on", 1, 1);
		$external_link_target = get_setting($preferences, "external link target", 1, 1);
		$search_results_preview_length = get_setting($preferences, "preview length", 1, 1);
		$line_ending = get_setting($preferences, "line ending", 1, 1);
		$show_source = get_setting($preferences, "show source", 1, 1);
		$show_source_to = get_setting($preferences, "show source to", 0, 1);
		$host_lookups = get_setting($preferences, "host lookups", 1, 1);
		$track_visitors = get_setting($preferences, "track visitors", 1, 1);
	} else {
		$field_marker = "";
		$log_traffic = "";
		$show_errors = "";
		$pre_render_files = "";
		$read_page_definitions_from = "";
		$read_files_from = "";
		$read_files_recursively = "";
		$require_authentication = "";
		$prevent_page_caching = "";
		$show_expanded_URLs = "";
		$split_page_names_on = "";
		$external_link_target = "";
		$search_results_preview_length = "";
		$line_ending = "";
		$show_source = "";
		$show_source_to = "";
		$host_lookups = "";
		$track_visitors = "";
	}
	
	
# do a bit of processing on the settings, and set some defaults in case preferences.ini or individual settings are absent

	if ($field_marker eq "") { $field_marker = "\#\#"; }
	
	if ($show_errors eq "") { $show_errors = "1"; }
	$show_errors = int($show_errors);

	if (($pre_render_files)&&($pre_render_files ne "1")&&($pre_render_files ne "0")) {
		$pre_render_files =~ s/, /,/g;
		@pre_render_files = split(",", $pre_render_files);
	}
	
	#if ($read_page_definitions_from eq "") { $read_page_definitions_from = "../contemplate/pages.txt"; }
	if ($read_page_definitions_from) {
		$read_page_definitions_from =~ s/, /,/g;
		@read_page_definitions_from = split(",", $read_page_definitions_from);
	} else {
		@read_page_definitions_from = ("data/pages.txt");
	}
	$read_page_definitions_from_count = @read_page_definitions_from;

	if ($read_files_from !~ /\.\.\/contemplate\//) { $read_files_from .= ", ../contemplate/"; }
	if ($read_files_from !~ /\.\.\/contemplate\/data\//) { $read_files_from .= ", ../contemplate/data/"; }
	if ($read_files_from !~ /\.\.\/content\//) { $read_files_from .= ", ../content/"; }
	if ($read_files_from !~ /\.\.\/templates\//) { $read_files_from .= ", ../templates/"; }
	if ($read_files_from =~ /^, (.*)/) { $read_files_from = $1; }
	if ($read_files_from) {
		$read_files_from =~ s/, /,/g;
		@read_files_from = split(",", $read_files_from);
	} else {
		@read_files_from = ();
	}
	$read_files_from_count = @read_files_from;

	if ($read_files_recursively eq "") { $read_files_recursively = "1"; }
	$read_files_recursively = int($read_files_recursively);

	if ($require_authentication) {
		$require_authentication =~ s/, /,/g;
		@require_authentication = split(",", $require_authentication);
	} else {
		@require_authentication = ();
	}
	$require_authentication_count = @require_authentication;

	if ($show_expanded_URLs eq "") { $show_expanded_URLs = "0"; }
	$show_expanded_URLs = int($show_expanded_URLs);

	if ($search_results_preview_length eq "") { $search_results_preview_length = 128; }

	if ($line_ending eq "") {
		$line_ending = "\n";
	} else {
		eval('$line_ending = '."\"$line_ending\";");
	}

	$show_source = int($show_source);

	if ($show_source_to) {
		$show_source_to =~ s/, /,/g;
		@show_source_to = split(",", $show_source_to);
		$show_source_to_count = @show_source_to;
	}

	if ($host_lookups eq "") { $host_lookups = "auto"; }
	
	if ($track_visitors eq "") { $track_visitors = "0"; }
	$track_visitors = int($track_visitors);


# check to see if the user is logged in via basic authentication
# to do: how can we access this in Perl?

	$login = ""; 


# generate an SQL-standard timestamp

	$now = time;
	($second, $minute, $hour, $day, $month, $year, $weekday, $yearday, $dst_flag) = localtime($now);
	$year += 1900;
	$month = sprintf("%02d",$month+1);
	$day = sprintf("%02d",$day);
	$hour = sprintf("%02d",$hour);
	$minute = sprintf("%02d",$minute);
	$second = sprintf("%02d",$second);
	$SQL_timestamp = "$year-$month-$day $hour:$minute:$second";
		

# write an HTTP header to prepare for text output

	$header .= "Content-type: text/html\r";
	if ($prevent_page_caching) {
  		# trick the browser into getting a new page every time
  		$header .= "Pragma: no-cache\r";
		$header .= "Cache-control: no-cache\r";
		$header .= "Expires: Mon, 07 May 1973 00:00:00 GMT\r";
	}
	if (($track_visitors)&&(!$cgi->cookie("Contemplate"))) {
		# if we want to track visitors and this visitor doesn't have a Contemplate cookie yet, generate a visitor ID and save it to the cookie (this assumes that if they have the cookie, it contains a visitor ID)
		$SQL_timestamp_stripped = $SQL_timestamp;
		$SQL_timestamp_stripped =~ s/\D//g;
		$visitor_ID = random_string(16)."-".$SQL_timestamp_stripped;
		@days = ("Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat");
		@months = ("Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec");
		$ten_years = time() + (60*60*24*365*10);
		($second, $minute, $hour, $day, $month, $year, $weekday, $yearday, $dst_flag) = localtime($ten_years);
		$weekday = @days[$weekday];
		$day = sprintf("%02d",$day);
		$month = @months[$month];
		$year += 1900;
		$expires = "$weekday, $day $month $year 00:00:00 GMT";
		$header .= "Set-Cookie: Contemplate=visitor_ID%3D$visitor_ID; Expires=".$expires."; Path=/; Version=1";
	}


# parse the URL to figure out the arguments for this page

	$page = $cgi->param('page');
	$requested_page = $cgi->param('requested_page');
	if ($page) {
		# look in the page definitions file(s)...
		$page_definition = "";
		for ($i=0; $i<$read_page_definitions_from_count; $i++) {
			$this_read_page_definitions_from = @read_page_definitions_from[$i];
			$this_read_page_definitions_from = relative_path($this_read_page_definitions_from, "contemplate");
			if ($this_read_page_definitions_from =~ /not authorized/i) {
				$output .= $this_read_page_definitions_from;
			} else {
				$page_definitions = get_file($this_read_page_definitions_from, "filesystem", 1);
				if ($file_missing) {
					$output .= $page_definitions; # $page_definitions will contain the error message generated by get_file()
				} else {
					$page_definition = get_setting($page_definitions, $page, 0, 0);
					if ($page_definition) { # CHECK FOR THE PRESENCE OF THE TEMPLATE VALUE ELSEWHERE
						# let any real-time args tag along for the ride
						$more_args = urldecode($ENV{'QUERY_STRING'});
						$more_args =~ s/&?page=[^&]+//;
						if ($more_args) { $page_definition .= "&".$more_args; }
						$page_definition =~ s/&+/&/g;
						if ($show_expanded_URLs) {
							# redirect to the page with the full URL visible
							$page_definition =~ s/&show_expanded_URLs=1//i;
							$page_definition .= "&requested_page=".$page;
							print $cgi->redirect("../contemplate/assembler.cgi?".$page_definition);
							exit;
						}
						$i = $read_page_definitions_from_count;
					}
				}
			}
		}
		if ($page_definition eq "") {
			$read_page_definitions_from_plural = ($read_page_definitions_from_count > 1) ? "s" : "" ;
			$output .= show_error("Can't find a definition for \"$page\" in the file$read_page_definitions_from_plural ".join(", ", @read_page_definitions_from).".");
			#$header = "Status: 404 Not Found\r".$header; # this seems to have the desired effect in IIS but not Apache
		}
	} else {
		# ...or just use the current URL
		$page_definition = urldecode($ENV{'QUERY_STRING'});
		if ($requested_page) { $page = $requested_page; }
	}
	
	# remove our own arguments from the page definition if present
	$page_definition =~ s/&formulator_fill=1//i;
	$page_definition =~ s/&show_info=1//i;

	# don't proceed from here unless we have a page definition
	if ($page_definition) {
	
		# create arrays from the argument values, so we can more easily refer to their individual parts
		@args = split("&", $page_definition);
		$args_count = @args;
		for ($i=0; $i<$args_count; $i++) {
			@pair = split("=", @args[$i]);
			$key = @pair[0];
			$key = lc($key); # argument names are hereby declared case-insensitive
			# skip args that have the same name as a preference setting, so that we can control which settings can be overridden
			if (",$settings_used," !~ /,$key,/) {
				$value = (@pair > 1) ? @pair[1] : "" ;
				# don't allow users to pass code through the URL arguments
				$value =~ s/</&lt;/i;
				$value =~ s/>/&gt;/i;
				if ($value =~ /,/) {
					@{$key} = split(",", $value);
				} else {
					if ($key !~ /template/) {
						@{$key}[0] = $value;
					} else {
						$template = $value;
					}
				}
			}
		}
		
	
# grab the text of the template file
	
		if ($template) {
			$template_text = relative_path($template, "templates");
			if ($template_text !~ /not authorized/i) { # if there's an error, leave the error in the template_text variable to display it
				$template_text = get_file($template_text, "auto", 1);
			}
		} else {
			$template_text = "<!--#embed args -->"; # allow display of one piece of content directly, with no template
		}
		$output .= $template_text;
		
	
# make the core processing recursive (e.g. process embed tags in embedded content)

		$i = 1;
		while (($output =~ /<!--\#\s*embed.*-->/is)||($output =~ /<!--\#\s*if.*-->/is)) {
		

# process any conditional tags to hide pieces of the template
	
			# first, find and number all if/else/end if sets
			@levels = ();
			$greatest = 0;
			while ($output =~ /<!--\#\s*end[ ]*if\D/is) {
				# which is next, an if, else, or end if?
				if ($output =~ /<!--\#\s*if\D(.*)/is) { $length_after_if = length($1); } else { $length_after_if = -1; }
				if ($output =~ /<!--\#\s*else\D(.*)/is) { $length_after_else = length($1); } else { $length_after_else = -1; }
				if ($output =~ /<!--\#\s*end[ ]*if\D(.*)/is) { $length_after_end_if = length($1); } else { $length_after_end_if = -1; }
				if (($length_after_if > $length_after_end_if)&&($length_after_if > $length_after_else)) {
					$greatest++;
					push(@levels, $greatest);
					$output =~ s/<!--\#\s*if(\D)/<!--\#if$greatest$1/i;
					#$output .= "<br>adding if $greatest";  # debug
				} elsif (($length_after_else > $length_after_if)&&($length_after_else > $length_after_end_if)) {
					$next = @levels[$levels-1];
					$output =~ s/<!--\#\s*else(\D)/<!--\#else$next$1/i;
					#$output .= "<br>adding else $next";  # debug
				} else {
					$next = pop(@levels);
					$output =~ s/<!--\#\s*end[ ]*if(\D)/<!--\#end if$next$1/i;
					#$output .= "<br>adding end if $next";  # debug
				}
			}
			
			# then evaluate and process each of them
			while ($output =~ /<!--\#\s*if(\d+)[ ]*\(([^\r<]*[^ ])\)[ ]*-->/is) {  # ignore extra spaces around "if" and expression
				$j = $1;
				$this_expression = $2;
				$native_expression = $this_expression;
				$native_expression =~ s/==/eq/g; $native_expression =~ s/!=/ne/g;
	
				# the variable symbol @ is optional in the template; add it here if it's missing
				if ($native_expression !~ /\x40[_\w]+\[/) { $native_expression =~ s/([_\w]+\[)/\x40$1/g; }
				
				# these variables don't exist internally as arrays; this fakes it so they're still available in conditional tags
				if (($split_page_names_on ne "")&&($native_expression =~ /\x40page\[\d+\]/ig)) {
					@page_array = split($split_page_names_on, $page);
					$native_expression =~ s/\x40page\[(\d+)\]/\x40page_array[$1]/ig;
				} else {
					$native_expression =~ s/\x40page\[\d*\]/\$page/ig;
				}
				$native_expression =~ s/\x40template\[\d*\]/\$template/ig;
				$native_expression =~ s/\x40page_definition\[\d*\]/\$page_definition/ig;
				
				# if an array (other than the special variables above) doesn't have an index, return every element of the array as a string
				$native_expression =~ s/\x40([_\w]+)\[\]/join(\",\", \@$1)/ig;

				# these keywords don't exist as arrays at all; this inserts their values as literal strings so they're available, too
				while ($native_expression =~ /\$*(env_\w+)/i) {
					$temp = "\"".process_tag_env($1)."\"";
					$native_expression =~ s/\$*env_\w+/$temp/;
				}
				while ($native_expression =~ /\$*(nav_\w+)/i) {
					$temp = "\"".process_tag_nav($1)."\"";
					$native_expression =~ s/\$*nav_\w+/$temp/;
				}

				# Perl doesn't seem to mind if we try to test a variable that hasn't been set, but we might as well find all the variables in the expression and test them in a safe way before evaluating it
				while ($native_expression =~ /\x40([_\w]+)\[([^\]])\]/i) {
					if ((@{$1})&&(@{$1} > $2)) { # to do: how can we check to make sure this variable is set? the exists() function throws an error 
						$temp = $field_marker.$1."[".$2."]"; # this ought to work right in the regex, but it doesn't
						$native_expression =~ s/\x40$1\[$2\]/$temp/ig;
					} else {
						$native_expression =~ s/\x40$1\[$2\]/""/ig;
					}
				}				
				$native_expression =~ s/$field_marker/\x40/ig;
				
				#$output .= $native_expression."<br>"; # debug
				$this_evaluates = eval($native_expression);
				if ($output =~ /<!--\#\s*if$j[ ]*\([^\r<]*[^ ]\)[ ]*-->(.*)<!--\#\s*end[ ]*if$j[ ]*-->/is) {
					# always show embed tags in the formulator
					$conditional_content = $1;
					if (($conditional_content =~ /<!--\#\s*embed/i)&&($formulator_fill)) { $this_evaluates = 1; }
				}
				if ($output =~ /<!--\#\s*else$j\D/i) {
					if ($this_evaluates) {
						$output =~ s/<!--\#\s*if$j[ ]*\(([^\r<]*[^ ])\)[ ]*-->//is; # remove the if tag...
						$output =~ s/<!--\#\s*else$j\D.*end[ ]*if$j[ ]*-->//is; # ...and from the else tag to the end if tag
					} else {
						$output =~ s/<!--\#\s*if$j\D.*else$j[ ]*-->//is; # remove from the if tag to the else tag...
						$output =~ s/<!--\#\s*end[ ]*if$j[ ]*-->//is; # ...and the end if tag
					}
				} else {
					if ($this_evaluates) {
						$output =~ s/<!--\#\s*if$j[ ]*\(([^\r<]*[^ ])\)[ ]*-->//is; # remove the if tag
						$output =~ s/<!--\#\s*end[ ]*if$j[ ]*-->//is; # remove the end if tag
					} else {
						$output =~ s/<!--\#\s*if$j\D.*end[ ]*if$j[ ]*-->//is; # remove from the if to the end if tag
					}
				}
			}
			
		
# find all the embed tags in the template and replace with temporary placeholders
		
			# slower: while ($output =~ /(.*?)<!--\#\s*embed\s*([^\r]*?[^ ])\s*-->(.*)/is) {  # ignore extra spaces before embed type, around "embed", and after attributes
			while ($output =~ /(<!--\#\s*embed\s*)([^\r]*?[^ ])(\s*-->)/is) {  # ignore extra spaces before embed type, around "embed", and after attributes
		
				#slower: $before_embed = $1; $this_embed = $2; $after_embed = $3;
				$this_embed_complete = $1.$2.$3;
				$this_embed = $2;
				$this_embed = lc($this_embed); # embed tags are hereby declared case-insensitive
				$this_embed_copy = $this_embed;
				$this_type = "";
				$this_embed_variable = "";
				$this_embed_strip = "";
				@this_embed_replace = ();
				$this_embed_fill_defer = "";
				$this_embed_special = "";
				
				if ($this_embed =~ /\s+variable/) {  # look for a variable attribute, which tells us to save the embed content as a variable rather than add it to the output
					$this_embed_variable = 1;
					$this_embed =~ s/\s+variable//;
				}
				
				if ($this_embed =~ /\s+strip/) {  # look for a strip attribute, which tells us to remove all markup from the embed content
					$this_embed_strip = 1;
					$this_embed =~ s/\s+strip//;
				}
				
				if ($this_embed =~ /\s+replace/) {  # look for a replace attribute, which tells us to do some search-and-replaces on the embed content
					@this_replace_array = ();
					while ($this_embed =~ /\s+replace\s+\/([^\/]*)\/([^\/]*)\//) {
						$search = $1; $replace = $2;
						push(@this_replace_array, $search);
						push(@this_replace_array, $replace);
						$this_embed =~ s/\s+replace\s+\/[^\/]*\/[^\/]*\///;
					}
					@this_embed_replace = @this_replace_array;
					# $output .= "this_replace_array: ".@embeds_replace[$i - 1]; # debug
				}
				
				if ($this_embed =~ /\s+fill_defer/) {  # look for a fill_defer attribute, which tells us put the formulator fill fields at the end of the file
					$this_embed_fill_defer = 1;
					$this_embed =~ s/\s+fill_defer//;
				}
				
				$output =~ s/<!--\#\s*embed\s*\Q$this_embed_copy\E\s*-->/-embed-$i-/ig; # this is global so that if the same embed tag appears more than once, we only have to parse it once
				#$output .= "<p>Found embed tag: $this_embed\r\r";  # debug
				

# process each embed tag and replace the placeholders with real content
		
			# show a warning if we have one of our special variable names; you can't name an embed tag with one of these names
		
				if (($this_embed eq "page")||($this_embed eq "template")||($this_embed eq "page_definition")) {
					$content = show_error("The tag name \"$this_embed\" is a reserved word<br>Hint: Choose a different name for this embed tag");
					$this_type = "ignore";
				}
				
		
			# get the args if we've hard-coded them into the tag
		
				if ($this_embed =~ /args[ ]*=[ ]*([^ ]*)/is) {
					$this_args = $1;
					while ($this_args =~ /(\w+)\[(\d*)\]/) {
						# process passthru/hard-coded combos
						$variable_name = $1;
						$variable_index = $2;
						# these variables don't exist internally as arrays; this fakes it so they're still available in passthru tags
						if (($variable_name eq "page")||($variable_name eq "template")||($variable_name eq "page_definition")) {
							if (($split_page_names_on ne "")&&($variable_name eq "page")) {
								if ($variable_index eq "") {
									# if no index is specified, return the whole page name
									$value = $page;
								} else {
									@page_array = split($split_page_names_on, $page);
									$value = @page_array[$variable_index];
								}
							} else {
								$value = ${$variable_name};
							}
						} else {
							if ($variable_index eq "") {
								# if no index is specified, return the whole array as a string
								$value = join(",", @{$variable_name});
							} else {
								$value = @{$variable_name}[$variable_index];
							}
						}
						$key = $variable_name."[".$variable_index."]";
						$this_args =~ s/\Q$key\E/$value/g;
					}
					$this_args =~ s/\+//g;
					$this_embed = "hardcoded"; # it doesn't really matter what we use for the name here
					@{$this_embed} = split(",", $this_args);
					$this_embed_special = 1;
				}
				
		
			# set the type to passthru if we have a square bracket (and remove the optional @ to avoid errors)
		
				if ($this_embed =~ /(\w+)\[(\d*)\]/) {
					$this_embed = $1;
					$this_embed_passthru_index = $2;
					$this_type = "passthru";
					if (($this_embed eq "page")||($this_embed eq "template")||($this_embed eq "page_definition")) {
						# treat these as hard-coded values that shouldn't be set in the formulator
						$this_embed_special = 1;
					}
				}
				
		
			# set the type to the name if we have a navigation or environment tag
		
				if (($this_embed =~ /nav_/i)||($this_embed =~ /env_/i)) {
					$this_type = $this_embed;
					$this_embed_special = 1;
				}
				
		
			# if we don't have any of the previously listed special types, the type is the first argument
		
				if (! $this_type) {
					if (@{$this_embed}) { # to do: how can we check to make sure this variable is set? the exists() function throws an error
						# we don't need to check isarray here because a string and an array can't have the same name in Perl
						$this_type = @{$this_embed}[0];
					} else {
						$this_type = "skip";
					}
				}
				
		
			# show popup menus rather than content if we're running from the formulator
		
				if ($formulator_fill) {
				
					# the first time through, strip any form tags from the template so they don't interfere with the popup menus
					if ($i == 1) {
						$output =~ s/<form.*?>//gi;
						$output =~ s/<\/form.*?>//gi;
					}
					
					# if this isn't a special kind of embed tag, make some menus, or an input box for passthru tags
					if (! $this_embed_special) {
					
						# don't display an input form for any tag more than once, even though it might appear in the template more than once
						$tag_names_flattened = join(",", @tag_names);
						if (",$tag_names_flattened," !~ /,$this_embed,/) {
						
							# if this is a passthru tag but a normal tag by this name appears later in the template, don't do anything now so we can fill out the argument using the menus
							if (!(($this_type =~ /passthr/)&&($output =~ /<!--\#\s*embed\s*$this_embed\s/is))) {  # ignore extra spaces before embed type, around "embed", and after attributes
							
								if ($this_type !~ /passthr/) {
								
									$fill_forms = "
									
										<b>".uc($this_embed)."</b>
										<a href=JavaScript:parent.inspect_content('$this_embed')
										 onmouseover='document.images.inspector_$this_embed.src=\"graphics/inspector_1.gif\"'
										 onmouseout='document.images.inspector_$this_embed.src=\"graphics/inspector_0.gif\"'><img 
										src=graphics/inspector_0.gif
										 width=12 height=12 hspace=2 border=0
										 name=inspector_$this_embed alt='Inspect'></a><br>
									
										<form name=contemplate_$this_embed>
									
											<select name=embed_type onchange=parent.set_type('$this_embed')>
												<option value=-1>Embed type...</option>
												<option value=file>File</option>
												<option value=field>Field</option>
												<option value=form>Form</option>
												<option value=random>Random</option>
												<option value=none>None</option>
											</select><br>
									
											<img src=../contemplate/graphics/spacer.gif width=1 height=2><br>
									
											<select name=embed_file onchange=parent.set_file('$this_embed')>
												<option value=0>(not needed)</option>
											</select><br>
									
											<img src=../contemplate/graphics/spacer.gif width=1 height=2><br>
									
											<select name=embed_field onchange=parent.set_field('$this_embed')>
												<option value=0>(not needed)</option>
											</select><br>
									
											<img src=../contemplate/graphics/spacer.gif width=1 height=2><br>
									
											<select name=embed_form>
												<option value=0>(not needed)</option>
											</select><p><br>
									
										</form>
										
									";
									
									if ($this_embed_fill_defer != 1) {
										# either display the popup menus right where the embedded content would be, and then stop parsing this tag...
										$content = $fill_forms;
										$this_embed_strip = "";
										@this_embed_replace = ();
										$this_embed_variable = "";
										$this_type = "ignore";
									} else {
										# ...or go ahead and parse this tag in place, and show the fill menus at the end of the page
										$output =~ s/<\/html>/$fill_forms<\/html>/is;
									}
				
									if ((@{$this_embed})&&(join(",", @{$this_embed}) ne "")) { # to do: how can we check to make sure this variable is set? the exists() function throws an error
										$this_embed_flattened = join(",", @{$this_embed});
										$stored_selections .= "parent.show_stored('$this_embed', '$this_embed_flattened');\r\t\t\t\t\t\t\t";				
									}
									
								} else {
								
									# process passthru values a little differently -- just show a text field for them
									
									$fill_forms = "
									
										<b>".uc($this_embed)."</b><br>
									
										<form name=contemplate_$this_embed>
									
											<input type=text size=24 name=embed_value><br>
									
										</form>
									
									";
									
									# display the passthru field at the top of the page, because those tags are likely to appear inside HTML tags
									# this also emphasizes their global scope
									$output =~ s/(<body[^>]*>)/$1$fill_forms/is;
					
									if ((@{$this_embed})&&(join(",", @{$this_embed}) ne "")) { # to do: how can we check to make sure this variable is set? the exists() function throws an error
										$this_embed_flattened = join(",", @{$this_embed});
										$this_embed_flattened =~ s/'/\\'/g; # in case the definition contains an apostrophe, which would break the JavaScript array setup code if not escaped
										$stored_selections .= "document.contemplate_$this_embed.embed_value.value = '$this_embed_flattened';\r\t\t\t\t\t\t\t";				
									}
				
								}
								
								push(@tag_names, $this_embed); # keep a list of all the embed tags we processed
						
							}
						
						}
		
					}
				
				}
		
		
			# open the content file, which we can assume to be the second argument
			
				if (($this_type eq "file")||($this_type =~ /field/)||($this_type =~ /random/)||($this_type =~ /form/)||($this_type =~ /menu/)) {
					$content_file = @{$this_embed}[1];
					$content_file = relative_path($content_file,"content");
					if ($content_file !~ /not authorized/i) {
						$content = get_file($content_file, "auto", 1);
					} else {
						$content = $content_file;
						$this_type = "error";
					}
					if ($file_missing) { $this_type = "error"; }
				}
		
		
			# get the desired piece of content...
			
			# ...for file
		
				if ($this_type =~ /file/) {
					# (no parsing necessary here)
				}
				
				
			# ...for field
				
				elsif ($this_type =~ /field/) {
					$file_type = @{$this_embed}[1]; # determines whether we parse as HTML or XML
					$content_field = @{$this_embed}[2];
					$content = get_field($content, $content_field, $file_type);
			   }
			   
			   
			# ...for random
		
				elsif ($this_type =~ /rand/) {
					$file_type = @{$this_embed}[1]; # determines whether we parse as HTML or XML
					@content_fields = get_field_names($content, $file_type);
					$count_fields = @content_fields;
					
					$content_field = @content_fields[int(rand($count_fields))];
					$content = get_field($content, $content_field, $file_type);       		
				}
				
		
			# ...for form
		
				elsif ($this_type =~ /form/) {
					$file_type = @{$this_embed}[1]; # determines whether we parse as HTML or XML
					$content_field = @{$this_embed}[2];
					$form_element = @{$this_embed}[3];
					$content = get_field($content, $content_field, $file_type);       		
					$content = get_form($content, $form_element);
				}
		
		
			# ...for search
				
				elsif ($this_type =~ /search/) {
					$search_string = $cgi->param('search_string');
					$search_files = $cgi->param('search_files');
					
					# skip if no search string entered
					if ($search_string ne "") {
					
						# if we haven't already loaded the page definitions file, do that now
						if (! $page_definitions) { $page_definitions = get_file(@read_page_definitions_from[0], "filesystem", 1); }
					
						# loop through each content file
						@content_files = get_recursive_listing("../content", $read_files_recursively, $search_files);
						$content_files_count = @content_files;
						for ($j=0; $j<$content_files_count; $j++) {
							$file_name = @content_files[$j];
							$file_text = get_file(relative_path($file_name, "content"), "filesystem", 0);
							#$output .= "Searching file ".relative_path($file_name, "content")."<p>\r\r"; # debug
			
							# if the file contains the search string...
							if ($file_text =~ /(.*?)($search_string)(.*)/is) {
							
								# ...create a preview of the found text...
								$results_preview = make_preview($1, $2, $3);
								
								# ...and count how many times it appears in the field
								$frequency_file = 0;
							
								# then get the field names in the file, and loop through each field
								@content_fields = get_field_names($file_text, $file_name);
								$content_fields_count = @content_fields;
								for ($k=0; $k<$content_fields_count; $k++) {
									$field_name = @content_fields[$k];
									$field_text = get_field($file_text, $field_name, $file_name);
									$field_name = @content_fields[$k]; # we need to reinstate $field_name, which get_field() inadvertently changes
								
									# if the field contains the search string...
									$field_text = strip_tags($field_text); # exclude text contained in a tag
									if ($field_text =~ /(.*?)($search_string)(.*)/is) {
										
										# ...create a preview of the found text...
										$results_preview = make_preview($1, $2, $3);
										
										# ...and count how many times it appears in the field
										$frequency_field = 0;
										while ($field_text =~ /$search_string/is) {
											$frequency_field++;
											$frequency_file++;
											$field_text =~ s/$search_string//i;
										}
			
										# look up the relevant page names, then add the preview and frequency to our results array
										$found_args = "field,$file_name,$field_name";
										@pages = get_page_names_containing($found_args, $page_definitions);
										$pages_count = @pages;
										for ($n=0; $n<$pages_count; $n++) {
											%results = (%results, "@pages[$n] : $results_preview" => $frequency_field);
										}
										
									}
								
								}
								
								# look up the relevant page names, then add the preview and frequency to our results array
								$found_args = "file,$file_name";
								@pages = get_page_names_containing($found_args, $page_definitions);
								$pages_count = @pages;
								for ($n=0; $n<$pages_count; $n++) {
									%results = (%results, "@pages[$n] : $results_preview" => $frequency_file);
								}
								
							}
							
						}
						
						# rank the search results by frequency
						@results_keys = sort { (($results{$b} <=> $results{$a}) || ($a cmp $b)) } keys %results; # this sorts first by frequency in descending order, then by name in ascending order
						@results_values = sort { int($b) <=> int($a) } values %results;
						
						#$content = "<table cellspacing=8 cellpadding=0 border=0>\r";  # these lines provide a different display option
						$content = "";
						$keys_count = @results_keys;
						for ($j=0; $j < $keys_count; $j++) {
							$this_page = substr(@results_keys[$j], 0, index(@results_keys[$j], " : "));
							$this_preview = substr(@results_keys[$j], index(@results_keys[$j], " : ") + 3);
							$this_frequency = @results_values[$j];
							#$content .= "<tr valign=top>\r<td class=body>$this_frequency</td>\r";
							#$content .= "<td class=body><a href=../contemplate/assembler.cgi?page=$this_page>contemplate/assembler.cgi?page=$this_page</a><br>\r<img src=../contemplate/graphics/spacer.gif width=1 height=4 alt=''><br>\r$this_preview\r</td>\r</tr>";
							if ($URL_rewriting_flag) {
								$content .= "<p class=\"search-results-preview\"><a href=\"../$URL_rewriting_flag/$this_page.html\" class=\"search-results-link\">$URL_rewriting_flag/$this_page.html</a><br>\r<img src=\"../contemplate/graphics/spacer.gif\" width=\"1\" height=\"4\" alt=\"\"><br>\r$this_preview ($this_frequency)</p>\r\r";
							} else {
								$content .= "<p class=\"search-results-preview\"><a href=\"../contemplate/assembler.cgi?page=$this_page\" class=\"search-results-link\">contemplate/assembler.cgi?page=$this_page</a><br>\r<img src=\"../contemplate/graphics/spacer.gif\" width=\"1\" height=\"4\" alt=\"\"><br>\r$this_preview ($this_frequency)</p>\r\r";
							}
						}
						#$content .= "</table>\r";
						
						if ($content) {
							$content = "<p class=\"search-results-heading\">$search_label_success \"$search_string\":</p>".$content;
						} else {
							$content = "<p class=\"search-results-heading\">$search_label_failure \"$search_string\"</p>";
						}
					} else {
						$content = "<p class=\"search-results-heading\">$search_label_none</p>";
					}
				}
		
		
			# ...for menu
			
				elsif ($this_type =~ /menu/) {
					$file_type = @{$this_embed}[1]; # determines whether we parse as HTML or XML
					$content_field = @{$this_embed}[2];
					$data_type = @{$this_embed}[3];
					$content = get_field($content, $content_field, $file_type);
					
					if ($data_type eq "graphics") { $data_type = "images"; } # this gives backwards compatibility from versions 1.1.0 to 1.2.2

					if ($content !~ /Assembler error/) {
	
						if ($data_type ne "html") { # for this type, we'll just return the complete content
					
							# strip out unnecessary tags and spaces to make this easier to parse
							$content =~ s/<\/li>//gi;
							$content =~ s/\r//g;
							$content =~ s/\s\s/ /g;
							
							# first, find and number all lists and items
							@levels = ();
							$greatest = 0;
							$limit = 0; # this is to prevent runaways while developing; we can delete it later
							while (($content =~ /<\/ul\D/is)&&($limit < 100)) {
								# which is next, a list item or a new child list?
								if ($content =~ /<li\D(.*)/is) { $length_after_item = length($1); } else { $length_after_item = -1; }
								if ($content =~ /<ul\D(.*)/is) { $length_after_list = length($1); } else { $length_after_list = -1; }
								if ($content =~ /<\/ul\D(.*)/is) { $length_after_end_list = length($1); } else { $length_after_end_list = -1; }
								if (($length_after_list > $length_after_item)&&($length_after_list > $length_after_end_list)) {
									$greatest++;
									push(@levels, $greatest);
									$content =~ s/<ul(\D)/<ul$greatest$1/i;
									#$output .= "<br>adding list $greatest";  # debug
								} elsif (($length_after_item > $length_after_list)&&($length_after_item > $length_after_end_list)) {
									$next = @levels[$levels-1];
									$content =~ s/<li(\D)/<li$next$1/i;
									#$output .= "<br>adding item $next";  # debug
								} else {
									$next = pop(@levels);
									$content =~ s/<\/ul(\D)/<\/ul$next$1/i;
									#$output .= "<br>adding end list $next";  # debug
								}
								$limit++;
							}
							
							# then work through each list from deepest to shallowest and create a string formatted like an array
							@child_lists = ();
							$child_lists[0] = ""; # we won't use the 0 index, but pad it now for consistency across languages
							for ($j=$greatest; $j>0; $j--) {
								if ($content =~ /<ul$j[^>]*>(.*)<\/ul$j>/is) {
									@child_lists[$j] = $1;
									$content =~ s/<ul$j[^>]*>.*<\/ul$j>/-child-list-$j-/is;
									@{@child_lists[$j]} = split(/<li[^>]*>/, @child_lists[$j]);
									for ($k=1; $k<@{@child_lists[$j]}; $k++) {
										#$output .= "processing list $j item $k<br>"; # debug
										if (@{@child_lists[$j]}[$k] =~ /-child-list-(\d+)-/) {
											$replace_index = $1;
											@{@child_lists[$j]}[$k] =~ s/-child-list-$replace_index-//;
										} else {
											$replace_index = 0;
										}
										if ($data_type eq "text") {
											@{@child_lists[$j]}[$k] = strip_tags(@{@child_lists[$j]}[$k]);
											@{@child_lists[$j]}[$k] =~ s/([\[\]\'\"])/\\$1/g; # escape characters that would break the array output
										}
										if ($data_type eq "images") {
											if (@{@child_lists[$j]}[$k] =~ /src\s*=\s*[\"\']?([^\"\'\s]+)[\"\'\s]/gi) {
												@{@child_lists[$j]}[$k] = $1;
											} else {
												@{@child_lists[$j]}[$k] = "";
											}
										}
										if ($data_type eq "images_width") {
											if (@{@child_lists[$j]}[$k] =~ /width\s*=\s*[\"\']?([^\"\'\s]+)[\"\'\s]/gi) {
												@{@child_lists[$j]}[$k] = $1;
											} else {
												@{@child_lists[$j]}[$k] = "";
											}
										}
										if ($data_type eq "images_height") {
											if (@{@child_lists[$j]}[$k] =~ /height\s*=\s*[\"\']?([^\"\'\s]+)[\"\'\s]/gi) {
												@{@child_lists[$j]}[$k] = $1;
											} else {
												@{@child_lists[$j]}[$k] = "";
											}
										}
										if ($data_type eq "images_alt") {
											if (@{@child_lists[$j]}[$k] =~ /alt\s*=\s*[\"\']?([^\"]+)[\"]/gi) {
												@{@child_lists[$j]}[$k] = $1;
												@{@child_lists[$j]}[$k] =~ s/([^\\])'/$1\\'/; # escape any apostrophes in the alt text (unless they are already escaped)
											} else {
												@{@child_lists[$j]}[$k] = "";
											}
										}
										if (($data_type eq "links")||($data_type eq "active")) {
											if (@{@child_lists[$j]}[$k] =~ /href\s*=\s*[\"\']?([^\"\'\s]+)[\"\'\s]/gi) {
												@{@child_lists[$j]}[$k] = $1;
											} else {
												@{@child_lists[$j]}[$k] = "";
											}
										}
										@{@child_lists[$j]}[$k] =~ s/^s*(.*)\s*$/$1/g; # no trim function?
										@{@child_lists[$j]}[$k] = "'".@{@child_lists[$j]}[$k]."'";
										if ($replace_index) {
											$temp = @child_lists[$replace_index];
											$temp =~ s/\[/\[@{@child_lists[$j]}[$k], /;
											@{@child_lists[$j]}[$k] = $temp;
										}
									}
									shift(@{@child_lists[$j]}); # the first element, anything before the first <li>, is junk, so throw it away
									@child_lists[$j] = "[".join(", ", @{@child_lists[$j]})."]";
								}
							}
							if (@child_lists > 1) {
								$content = @child_lists[1];
							} else {
								$content = "[]";
							}
							
							# if the data type is active, loop through the links array and find the active link (it would be nicer to catch this as we build the arrays, but at that point we don't know what position the active element will eventually hold in the multidimensional array)
							if ($data_type eq "active") {
								$links = $content;
								$links =~ s/^\[/(/;
								$links =~ s/\]$/)/;
								eval('@links = '.$links.';');
								$array_path = "";
								for ($j=0; $j<@links; $j++) {
									# in (1,2,(a,b)), (a,b) isn't recognized as a separate list
									# but in (1,2,[a,b]), [a,b] gets munged when passed to a function like join or a subroutine like join_recursive
									# so this code will work with two-dimensional lists, but we can't find matches deeper than that because they get munged by the join
									# to see an example, create a three-level list and enable the debug line below
									if (@{@links[$j]}) {
										$element_flattened = join(",", @{@links[$j]});
									} else {
										$element_flattened = join(",", @links[$j]);
									}
									#$output .= "checking element $j of ".@links.": ".$element_flattened."<br>"; # debug
									if ($element_flattened =~ /\b$page\b/i) {
										#$output .= "element $j is a match<br>"; # debug
										$array_path .= $j;
										if ($element_flattened =~ /,/) {
											$array_path .= ",";
											if (@{@links[$j]}) {
												@links = @{@links[$j]};
											} else {
												@links = @links[$j];
											}
											$j = 0;
										} else {
											break; # we found it!
										}
									}
								}
								$content = "[".$array_path."]";
							}
							
						}

					}
				}
		
		
			# ...for navigation
			
				elsif (($this_type =~ /nav_current/)||($this_type =~ /nav_total/)||($this_type =~ /nav_next/)||($this_type =~ /nav_prev/)||($this_type =~ /nav_first/)||($this_type =~ /nav_last/)) {
					$content = process_tag_nav($this_type);						
				}
		
		
			# ...for environment variable
			
				elsif ($this_type =~ /env_/) {
					$content = process_tag_env($this_type);
										
					# uncomment this to see all available variables
					#foreach $envbit (%ENV) { $output .= "$envbit<br>"; }  # debug
				}
		
		
			# ...for passthrough/passthru
			
				elsif ($this_type =~ /passthr/) {
					# these variables don't exist internally as arrays; this fakes it so they're still available in passthru tags
					if (($this_embed eq "page")||($this_embed eq "template")||($this_embed eq "page_definition")) {
						if (($split_page_names_on ne "")&&($this_embed eq "page")) {
							if ($this_embed_passthru_index eq "") {
								# if no index is specified, return the whole page name
								$content = $page;
							} else {
								@page_array = split($split_page_names_on, $page);
								if (@page_array > $this_embed_passthru_index) {
									$content = @page_array[$this_embed_passthru_index];
								} else {
									$content = "";
								}
							}
						} else {
							$content = ${$this_embed};
						}
					} else {
						if ($this_embed_passthru_index eq "") {
							# if no index is specified, return the whole array as a string
							$content = join(",", @{$this_embed});
						} else {
							if (@{$this_embed} > $this_embed_passthru_index) {
								$content = @{$this_embed}[$this_embed_passthru_index];
							} else {
								$content = "";
							}
						}
					}
				}
		
		
			# ...or for nothing!
				
				elsif ($this_type =~ /skip/) {
					$content = "<!-- Embed tag \"$this_embed\" not parsed because of blank or too few arguments in URL -->";
				}
		
				elsif ($this_type =~ /ignore/) {
					# do nothing at all
				}
		
				elsif ($this_type !~ /error/) {
					$content = show_error("Can't understand the data type \"$this_type\"<br>Available types: file, field, random, form, search, nav_current, nav_total, nav_next, nav_previous, nav_first, nav_last, env_*, or passthru$check_embed");
				}
				
				
			# skip the rest of the parsing if this tag generated an error
			
				if ($content !~ /Assembler error/) {
				
				
				# remove all markup from the embed content if the strip attribute was included in the embed tag
				
					if ($this_embed_strip) {
						$content = strip_tags($content);
					}
				   
				   
				# do some search-and-replaces if the replace attribute was included in the embed tag
				
					if (@this_embed_replace) {
						$replace_count = @this_embed_replace / 2;
						for ($j=0; $j < $replace_count; $j++) {
							$search = @this_embed_replace[$j*2];
							$replace = @this_embed_replace[($j*2)+1];
							eval('$content =~ s/'.$search.'/'.$replace.'/g;');
						}
					}
				
				
				}
			
			
			# add the new content to the output...
			
				if (! $this_embed_variable) {
		
					# replace all instances normally, but only the first instance in the formulator
					if (!$formulator_fill) {
						$output =~ s/-embed-$i-/$content/g;
					} else {
						$output =~ s/-embed-$i-/$content/;
						# delete the extra embed placeholders that we didn't use in the formulator
						$output =~ s/-embed-$i-//g;
					}
				
				} else {
				
		
			# ...or just save it as a variable and delete the embed placeholder
				
					push(@embed_variable_keys, $this_embed);
					push(@embed_variable_values, $content);
					
					$output =~ s/-embed-$i-//g;
		
				}
				
		
			# finally, start over with the next embed
			
				$last_content = $content;
				$content = "";
				$i++;
			
			}
	
		}


# add formulator data to the bottom of the document (we can't do this until we're done with recursive parsing)

	if ($formulator_fill) {
		
		if (@tag_names) { # don't bother with all this processing if there are no fillable tags
		
			$content_folder_only = 1; # we'll tell JavaScript whether we're using any nonstandard file locations
			@content_files = ();
			for ($j=0; $j<@read_files_from; $j++) {
				$this_directory = @read_files_from[$j];
				if (($this_directory !~ /\.\.\/contemplate/)&&($this_directory !~ /\.\.\/templates/)) {
					if ($this_directory !~ /\.\.\/content/) { $content_folder_only = 0; }
					@this_files = get_recursive_listing($this_directory, $read_files_recursively, "");
					for ($k=0; $k<@this_files; $k++) {
						$this_file = @this_files[$k];
						push(@content_files, $this_directory.$this_file);
					}
				}
			}
			$file_fields = "";
			$field_forms = "";
			
			# build an array of fields for each content file...
			for ($j=0; $j<@content_files; $j++) {
				$this_file_name = @content_files[$j];
				$this_file_nopunctuation = $this_file_name;
				$this_file_nopunctuation =~ s/\./_/g;
				$this_file_nopunctuation =~ s/\//_/g;
				$this_file_nopunctuation =~ s/-/_/g;
				$this_file_nopunctuation =~ s/ /_/g;
				$this_file_nopunctuation =~ s/^_*(.*)_*$/$1/g;
				$this_file_text = get_file($this_file_name, "filesystem", 0);
				@this_file_fields = get_field_names($this_file_text, $this_file_name);
				if (@this_file_fields) {
					$this_file_fields = join("', '", @this_file_fields);
					$file_fields .= "parent.".$this_file_nopunctuation."_fields = Array('$this_file_fields'); // file $j\r\t\t\t\t\t\t\t";
				}
				
				# ...and an array of forms for each field
				if (@this_file_fields) { @this_file_fields = split("', '", $this_file_fields); }
				for ($k=0; $k<@this_file_fields; $k++) {
					$this_field_name = @this_file_fields[$k];
					$this_field_nopunctuation = $this_field_name;
					$this_field_nopunctuation =~ s/\./_/g;
					$this_field_nopunctuation =~ s/\//_/g;
					$this_field_nopunctuation =~ s/-/_/g;
					$this_field_nopunctuation =~ s/ /_/g;
					$this_field_nopunctuation =~ s/^_*(.*)_*$/$1/g;
					$this_field_text = get_field($this_file_text, $this_field_name, $this_file_name);
					@this_field_forms = get_form_names($this_field_text);
					if (@this_field_forms) {
						$this_field_forms = join("', '", @this_field_forms);
						$field_forms .= "parent.".$this_file_nopunctuation."_".$this_field_nopunctuation."_forms = Array('$this_field_forms'); // form $j:$k\r\t\t\t\t\t\t\t";
					}
				}
			}
			
			$tag_names_javascript = join ("', '", @tag_names);
			$content_files_javascript = join ("', '", @content_files);
	
			$fill_data = "
			
				<script language=JavaScript>
			
					function load_menu(menu_name, menu_title, values) {
						clear_menu(menu_name, menu_title);
						menu_name.options[0] = new Option(menu_title, '0');
						if (values.length) {
							for (i=0; i < values.length; i++) {
								this_value = values[i];
								if (menu_title == 'File...') {
									if (parent.content_folder_only) {
										// remove the '../content/' if all the files are in that folder
										this_value_display = this_value.substring(11, this_value.length);
									} else {
										// just remove the '../' otherwise
										this_value_display = this_value.substring(3, this_value.length);
									}
								} else {
									this_value_display = this_value;
								}
								menu_name.options[i+1] = new Option(this_value_display, this_value);
							}
						} else {
							menu_name.options[1] = new Option('(none available)', '0');
						}
					}
					
					function clear_menu(menu_name, menu_title) {
						menu_name.options[0] = new Option(menu_title, '0');
						for (i=menu_name.options.length; i > 1; i--) {
							menu_name.options[i-1] = null;
						}
					}
	
					parent.tag_types = Array('file', 'field', 'form', 'random', 'none');
					parent.tag_names = ('$tag_names_javascript') ? Array('$tag_names_javascript') : Array() ;
					parent.content_files = ('$content_files_javascript') ? Array('$content_files_javascript') : Array() ;
					
					parent.content_folder_only = $content_folder_only;
	
					$file_fields
					
					$field_forms
					
					$stored_selections
			
					parent.contemplate_toolbar.location = 'formulator/toolbar.html?step=4'; // the formulator waits for this code to reload the toolbar

				</script>
				
			";
			
		} else {
		
			# if there are no fillable tags, advance the formulator toolbar to the next step
			$fill_data = "
				<script language=JavaScript>
					parent.tag_names = Array();
					parent.contemplate_toolbar.location = 'formulator/toolbar.html?step=5'; // the formulator waits for this code to reload the toolbar
				</script>
			"; 
			
		}
		
		$output =~ s/<\/html>/$fill_data<\/html>/is; # insert the JavaScript at the end of the file
		
	}


# merge argument values back together, since other scripts won't expect them to be arrays
	
		for ($i=0; $i<$args_count; $i++) {
			$pair = split("=", $args[$i]);
			$key = @pair[0];
			if (",$settings_used," !~ /,$key,/) {
				$value = (@pair > 1) ? @pair[1] : "" ;
				${$key} = $value;
			}
		}
		
	
# put content into any embed tag names marked with the variable attribute (we do this after merging argument variables in case they have the same names)
	
		$embed_variable_keys_count = @embed_variable_keys;
		for ($i=0; $i<$embed_variable_keys_count; $i++) {
			${@embed_variable_keys[$i]} = @embed_variable_values[$i];
		}


# set external link targets if requested in prefs file
	
		if ($external_link_target) {
			# remove any quotes from the external link target, and we'll make it match the quoting style of the href value
			$external_link_target =~ s/[\"\']//g;
			# add our default target to every anchor tag
			$output =~ s/<a href=(\\?[\'\"]?)(\w+:\/\/[^>]*)>/<a href=$1$2 target=$1$external_link_target$1>/g;
			# then remove it from every tag that already had one
			$output =~ s/<a href=(\\?[\'\"]?)(\w+:\/\/[^>]*target=[^>]*) target=[^>]*>/<a href=$1$2>/g;
		}
	
	
	} # end of the "if $page_definition" check


# return the parsed text as an HTML file

	$header =~ s/\r/\n/g; # although we're using Mac line breaks during parsing, the header needs to be Unix
	print $header."\n";
	
	# change to the preferred line ending set in preferences.ini
	$output =~ s/\r/$line_ending/g;

	print $output;
	
	# if requested, display the output in a numbered, pre-rendered format so we can see where an error occurred
	# but only allow this from IP addresses listed in the preferences file
	$show_source_approved = 0;
	if (@show_source_to) {
		for ($i=0; $i < $show_source_to_count; $i++) {
			if ($ENV{'REMOTE_ADDR'} eq @show_source_to[$i]) { $show_source_approved = 1; }
		}
	}
	if (($show_source)&&($show_source_approved)) {
		$output_numbered = $output;
		$output_numbered =~ s/</&lt;/g;
		$output_numbered =~ s/>/&gt;/g;
		@output_numbered = split($line_ending, $output_numbered);
		$output_numbered_count = @output_numbered;
		for ($i=0; $i<$output_numbered_count; $i++) {
			@output_numbered[$i] = "<tr valign=top><td><font face=geneva,verdana,sans-serif color=#999999>".($i+1)."</font></td><td>&nbsp;</td><td><font face=geneva,verdana,sans-serif>".@output_numbered[$i]."</font></td></tr>";
		}
		$output_numbered = join($line_ending, @output_numbered);
		print "<html>\r<body bgcolor=#FFFFFF>\r<p>\r<table>\r$output_numbered\r</table>\r</p>\r</body>\r</html>\r";
	}


# log this hit in the traffic reports

	if (($log_traffic)&&(!$skip_rendering)&&(!$formulator_fill)) {
	
		$status = ($ENV{'REDIRECT_STATUS'}) ? $ENV{'REDIRECT_STATUS'} : "" ;

		if (($request =~ "assembler.cgi")||($URL_rewriting_flag)||($status eq "200")) {
			# if we recognize the request as a call to the assembler, display either the page name or the expanded URL
			if (!$page) { $page = $page_definition; }
		} else {
		 	# ... otherwise, we must have been redirected to here, so show the original request with the error code
		 	$page = substr($request,1); # the substr removes the initial slash
		 	$error_log = ($status) ? $status : "Unknown" ;
		}
		
		# catch the other environment variables now, or bring them in from the wrapper when we're using that (Perl only)
		$from_wrapper = $cgi->param('from_wrapper');
		if ($from_wrapper) {
			$IP_address = $cgi->param('IP_address');
			$host = $cgi->param('host');
			$user_agent = $cgi->param('user_agent');
			$referrer = $cgi->param('referrer');
			$page = substr($page, 0, index($page, "&from_wrapper"));
		} else {
			$IP_address = $ENV{'REMOTE_ADDR'};
			$host = ($ENV{'REMOTE_HOST'}) ? $ENV{'REMOTE_HOST'} : $IP_address ; # if we knew that all servers substituted the IP address for us when the host is missing, then we could remove the conditional here
			$user_agent = $ENV{'HTTP_USER_AGENT'};
			$referrer = $ENV{'HTTP_REFERER'};
		}

		# find out whether we need to do our own host lookup
		if ($host_lookups eq "auto") {
			$host_lookups = ($host eq $IP_address) ? "on" : "off" ;
		}
		if ($host_lookups eq "on") {
			eval 'use Net::hostent';
			if (!$@) {
				if (($host !~ /192\.168\.0\./)&&($host !~ /10\.0\.1\./)&&($host !~ /127\.0\.0\./)) {
					$host = pack("C4", split(".", $host));
					$host = (gethostbyaddr $host, 2); #->name;
				}
			}
		}

		# add the visitor ID to each entry if we're using cookies to track visitors
		if ($track_visitors) {
			if ($visitor_ID eq "") {
				# get the visitor ID from the cookie if we didn't just generate one on this page
				# if we ever use the cookie to store other data, we can move this code to the top the script so that it's available to the rest of the script; right now it's here so that we can only run it when we need it
				$cookie_data = ();
				$cookie_data_raw = $cgi->cookie("Contemplate");
				@cookie_data_raw = split("&", $cookie_data_raw);
				for ($i=0; $i<@cookie_data_raw; $i++) {
					@cookie_data_item = split("=", @cookie_data_raw[$i]);
					@cookie_data[@cookie_data_item[0]] = @cookie_data_item[1];
				}
				$visitor_ID = @cookie_data["visitor_ID"];
			}
			$visitor_ID_with_tab = "$visitor_ID\t";
		} else {
			$visitor_ID_with_tab = "";
		}
		
		@SQL_timestamp_split = split("-", $SQL_timestamp);
		$year = @SQL_timestamp_split[0];
		$month = @SQL_timestamp_split[1];
		$traffic_file = absolute_path("data/reporter/".$year."-".$month.".txt", "filesystem");
		$error_file = absolute_path("data/reporter/".$year."-".$month."-errors.txt", "filesystem");
		
		if (!$error_log) {
			open (trafficfile, ">>$traffic_file") or
			$output_two .= "<!-- Can't open the traffic data file $traffic_file: $! -->\r";
			unless ($output_two =~ /data/) {
				print trafficfile "$SQL_timestamp\t$visitor_ID_with_tab$IP_address\t$host\t$page\t$user_agent\t$referrer\r";
				close trafficfile;
			}
		} else {
			open (errorfile, ">>$error_file") or
			$output_two .= "<!-- Can't open the error file $error_file: $! -->\r";
			unless ($output_two =~ /Can't open/) {
				print errorfile "$SQL_timestamp\t$visitor_ID_with_tab$IP_address\t$host\t$page\t$user_agent\t$referrer\t$error_log\r";
				close errorfile;
			}
		}
		
		print $output_two;
		
	}


# we're done!

	exit;
